/*
  Filename: paclock.c  
  Synopsys: PA clock widget based on timux library
    
 	                      Copyright (C) 2001  
	         David Gmez Espinosa <david@pleyades.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "pabar.h"
#include "paclock.h"
#include "timux.h"


struct sPAClock {
    Window window;            /* The Puto Amo Clock window */
    GC gc;                    /* GC for the PAClock */      
    GC maingc;                /* GC for the main display */ 
    XFontStruct *font;        /* Font for the PAClock */
    struct tm *thetime;       /* The Time (tm) */
    unsigned int baseline;    /* Baseline for PAClock text*/
    jobid job;
}; 

#define PAClockBytes (sizeof(struct sPAClock))
#define PACLOCK_EVMASK (ExposureMask)

PAClock thePAClock=NULL;
TiMux paclock_timer;
Display *clockdpy;  /* PAClock own display */


static void paclock_handler (XEvent *, void *);
static void paclock_update (void *);
static void paclock_draw (Display *, GC);

static unsigned short int reference_count=0;


void paclock_init (void) 
{
    time_t seconds;
    struct sigaction sa;
    char *tmpvalue;
    
    if (reference_count++) return;

    thePAClock=(malloc(PAClockBytes));
    if (!thePAClock) die("Not enough memory");

    /* PAClock update function is called from inside a signal handler, so we
       need a different display to be used exclusively by the handler. If
       not, doing Xlib calls will trash the protocol buffer when the signal is
       handled during a Xlib function. */
   
    clockdpy=XOpenDisplay(display_name);
    if (!clockdpy) die("Cannot open display");

    time(&seconds);
    thePAClock->thetime=localtime(&seconds);

    sa.sa_handler=SIG_IGN;
    sigaction(SIGALRM, &sa, NULL);
    
    paclock_timer=timux_create();

    thePAClock->gc=XCreateGC(clockdpy, ROOT, 0, NULL);
    if (thePAClock->gc==None) BUG();
    thePAClock->maingc=XCreateGC(display, ROOT, 0, NULL);
    if (thePAClock->maingc==None) BUG();

    XSetForeground(display,thePAClock->maingc,x_color("clock_border"));	 
    XSetForeground(display,thePAClock->gc,x_color("clock_border"));	 
    
    tmpvalue=cfgpool_getstring(pawmpool,"clock_font");
    if (!tmpvalue) 
	die("Needed clock font in conf file!");
    thePAClock->font=XLoadQueryFont(display,tmpvalue);
    free(tmpvalue);

    /* Get the baseline for centering the text */
    thePAClock->baseline=((PACLOCKH-(thePAClock->font->ascent+thePAClock->font->descent))/2)+thePAClock->font->ascent;

    XSetFont(display, thePAClock->maingc, thePAClock->font->fid);
    XSetFont(clockdpy, thePAClock->gc, thePAClock->font->fid);

    thePAClock->window=XCreateSimpleWindow (display, pabar_getwindow(), 
		                            DWIDTH-PACLOCKW-1,(PABARH-PACLOCKH)/2, 
		    		     	    PACLOCKW, PACLOCKH , 0,
                                            BLACK, x_color("frame_base_color"));
    if (thePAClock->window==None) BUG();

    XStoreName(display, thePAClock->window,"paclock");

    XSelectInput(display, thePAClock->window, PACLOCK_EVMASK);
    x_addsloot(thePAClock->window, paclock_handler, NULL);

    XMapRaised(display, thePAClock->window);

    /* Add a timux job to update the clock */
    thePAClock->job=timux_addjob(paclock_timer,paclock_update,NULL,2000);
    
}



void paclock_done (void) 
{
    if (--reference_count) return;

    timux_deljob(paclock_timer,thePAClock->job);
    timux_delete(paclock_timer);
    
    XFreeFont(display,thePAClock->font);
    XFreeGC(clockdpy, thePAClock->gc);
    XFreeGC(display, thePAClock->maingc);

    x_delsloot(thePAClock->window, paclock_handler, NULL);    
    XUnmapWindow(clockdpy, thePAClock->window);
    XDestroyWindow(clockdpy, thePAClock->window);

    free(thePAClock);

}


static void paclock_handler(XEvent *xevent, void *widget) 
{
    ASSERT(xevent);

    widget=NULL;

    switch(xevent->type) {

    case Expose:
        GOT(Expose, PAClock);
	if ((xevent->xexpose).count) return;
	paclock_draw(display,thePAClock->maingc);
        break;

    default:
	SAY("Event %d on PAClock %ld",xevent->type, thePAClock->window);
    }
}


static void paclock_update(void *noparam)
{
 
    time_t seconds;

    if (noparam) BUG();
 
    time(&seconds);
    thePAClock->thetime=localtime(&seconds);

    paclock_draw(clockdpy,thePAClock->gc);

    XFlush(clockdpy);

    /* Add the job again... */
    thePAClock->job=timux_addjob(paclock_timer,paclock_update,NULL,1000);

}


void paclock_draw(Display *dpy, GC gc)
{
	
   char clocktext[7];
   
   /* Draw time */
   strftime(clocktext,sizeof(clocktext),"%H:%M",thePAClock->thetime);
   XClearWindow(dpy,thePAClock->window);
   XDrawString(dpy,thePAClock->window,gc,
	       (int) (PACLOCKW-XTextWidth(thePAClock->font, clocktext, strlen(clocktext)))/2,
               (int) thePAClock->baseline,
               clocktext,(int) strlen(clocktext));

   /* Draw clock outline */
   XDrawRectangle(dpy,thePAClock->window,gc,0,0,PACLOCKW-1,PACLOCKH-1);

}

