/*
  Filename: paicon.c  
  Synopsys: An Icon class
    
 	                      Copyright (C) 2003, 2004 
	         David Gmez Espinosa <david@pleyades.net>

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <X11/extensions/shape.h>
#include "paicon.h"


#define ICON_MARGIN 5       /* Margin between icon and icon text */
#define TEXT_TOPMARGIN 2
#define TEXT_SIDEMARGIN 7
/* Max icon text width */
#define TEXTI_W (icon->image->w*2)  
/* Max icon text height */
#define TEXTI_H ((class.font->ascent+class.font->descent+TEXT_TOPMARGIN)*2) 
#define TEXTW (TEXTI_W+TEXT_SIDEMARGIN)
#define TEXTH TEXTI_H



struct sPAIcon {
    Window iconwindow;         
    Window textwindow;                  
    Window layer;
    int icon_x,icon_y;		        /* Position of the icon */
    int text_x,text_y;                  /* Position of the icon text */
    xImage image;                       /* Image data */
    char *icon_file;                    /* Icon xpm file name */
    char *icon_text;                    /* Text showing in screen */
    char **app;	                 	/* App & params to execute */
    char *line1, *line2;
};

static struct {
    XFontStruct *font;
    GC gc;
    unsigned long textcolor;
    unsigned long background;
    Cursor iconcursor;
    Bool moving;
} class;

#define PAIconBytes (sizeof(struct sPAIcon))

#define PAICON_EVMASK (ButtonPressMask|ButtonReleaseMask|ExposureMask) 
#define PAICONL_EVMASK (ExposureMask)
#define PAICON_TEXTEVMASK (ButtonPressMask|ExposureMask)


static unsigned short int reference_count=0;

static void paicon_handle (XEvent *, void *);
static void paicon_ButtonPress (XButtonPressedEvent *, PAIcon );
static void paicon_Expose (XExposeEvent *, PAIcon);
static void paicon_exec (PAIcon);

/* Helpers */
static void paicon_draw (PAIcon);
static void paicon_drawtext(PAIcon);
static void paicon_getlines(PAIcon);


static void paicon_handle (XEvent *xevent, void *widget) {

    PAIcon icon=NULL;

    ASSERT(xevent);
    ASSERT(widget);
        
    icon=(PAIcon)widget;

    switch(xevent->type) {
    case ButtonPress:
        GOT(ButtonPress, PAIcon);
        paicon_ButtonPress(&(xevent->xbutton), icon);
        break;
    case ButtonRelease:
        GOT(ButtonRelease, PAIcon);
        break;
    case Expose:
        GOT(Expose, PAIcon);
        paicon_Expose(&(xevent->xexpose), icon);
        break;
    }
}


static void paicon_ButtonPress (XButtonPressedEvent *event, PAIcon icon) 
{

    int x1=0, y1=0, x2=0, y2=0;
    int xoff=0, yoff=0;
    long evmask;
    unsigned int tmpmask;
    Window root=None;
    Window window=None;
    Bool moved=False;
    
    ASSERT(event);
    ASSERT(icon);    

    if (event->button!=Button1) return;
    
    if (XQueryPointer(display, icon->iconwindow, &root, &window,
                      &x1, &y1, &x2, &y2, &tmpmask)==False) BUG();

    /* Determine pointer offset */
    xoff=x1-icon->icon_x;
    yoff=y1-icon->icon_y;
	    
    /* Special event mask for doing the move */
    evmask=ButtonReleaseMask|PointerMotionMask;

    /* Grab the pointer while moving the icon */
    XGrabPointer(display, ROOT, True, (unsigned int)evmask,
                 GrabModeAsync, GrabModeAsync, None,
                 None, CurrentTime);

    /* Get expose events for updating the screen while the thing is grabbed */
    evmask|=ExposureMask;
    
    while (1) {  /* Loops 'til button released */

	XEvent xevent;

        XMaskEvent(display, evmask, &xevent);

        switch (xevent.type) {

        case MotionNotify: {   
            XMotionEvent mevent=xevent.xmotion;
	    XEvent foo;

            if (!class.moving) return;  /* No moving, no work */
	    if (event->window==icon->textwindow) break;
    
	    if (!moved) {
	        /* Unmap the text while moving the icon */
		XDefineCursor(display, icon->layer, class.iconcursor);
	        XUnmapWindow(display,icon->textwindow);
		
	    }

            moved=True;

	    /* Discard excess of Motion events */
	    while (XCheckTypedEvent(display, MotionNotify,&foo));

            x1=mevent.x_root-xoff;
            y1=mevent.y_root-yoff;

	    /* Move the icon */
	    paicon_setpos(icon,x1,y1);
	    XMapRaised(display,icon->iconwindow);
	    paicon_draw(icon);
	    XRaiseWindow(display, icon->textwindow);
	    XMapRaised(display, icon->layer);
	    
            break;
        }
			   
       case Expose: {
            x_dispatch(&xevent);			 
	    break;
	}

       case ButtonRelease: {  /* End the moving */
			
	    XButtonPressedEvent *mevent=&(xevent.xbutton);
	   
	    /* Make visible again the icon text window */
	    XMapWindow(display, icon->textwindow);
	    paicon_drawtext(icon);

	    XUndefineCursor(display, icon->layer);
            XUngrabPointer(display, CurrentTime);

	    /* Run the program if there hasn't been movement. If the button
	       release ocurred outside the text window, do nothing. This
	       hell-of-a-condition is here because the fucking ButtonRelease
	       is fucking reported relative to the fucking root window. Fuck */
	    if ( (event->window==icon->layer && !moved) ||
		    (event->window==icon->textwindow && !moved &&
		     ( ((mevent->x > icon->text_x) && (mevent->x < icon->text_x+TEXTW)) &&
		       ((mevent->y > icon->text_y) && (mevent->y < icon->text_y+TEXTH)) )))
		
	        paicon_exec(icon);

            return;
       }
     }
    }
}


static void paicon_Expose (XExposeEvent *event, PAIcon icon) {

    ASSERT(event);
    ASSERT(icon);

    if (event->count) return;  /* More exposes to come */

    paicon_draw(icon);
    paicon_drawtext(icon);
}




/* Public functions */

void paicon_init (void) 
{
    char *tmpvalue;
    
    if (reference_count++) return;

    tmpvalue=cfgpool_getstring(pawmpool,"paicon_move");
    class.moving=(strcmp(tmpvalue,"On")?False:True); free(tmpvalue);

    class.gc=XCreateGC(display, ROOT, 0, NULL);
    if (class.gc==None) BUG();

    tmpvalue=cfgpool_getstring(pawmpool,"icons_font");
    class.font=XLoadQueryFont(display,tmpvalue); free(tmpvalue);
    
    class.textcolor=x_color("icon_text");
    class.background=x_color("background");

    class.iconcursor=XCreateFontCursor(display, XC_fleur);

    XSetForeground(display,class.gc,class.textcolor);
    XSetFont(display,class.gc,class.font->fid);
}



void paicon_done (void)
{
    if (--reference_count) return;
    
    XFreeGC(display, class.gc);
    XFreeFont(display, class.font);
    XFreeCursor(display, class.iconcursor);

}


PAIcon paicon_create (xImage image)
{
    PAIcon icon=NULL;

    ASSERT(image);

    icon=malloc(PAIconBytes);
    if (!icon) die("Not enough memory");

    icon->image=image;
    
    icon->iconwindow=XCreateSimpleWindow(display, ROOT, 0, 0, image->w, image->h,
                                         0, BLACK, BLACK);
    if (icon->iconwindow==None) BUG();

    icon->textwindow=XCreateSimpleWindow(display, ROOT, 0, 0, 
	                                 TEXTW, TEXTH,
	    				 0, BLACK, class.background);
    if (icon->textwindow==None) BUG();
    
    icon->layer=XCreateWindow(display, ROOT, 0, 0, image->w, image->h, 0, 0,
	                      InputOnly, VISUAL, 0, NULL);
    if (icon->layer==None) BUG();
    
    /* Got mask, apply it to the image */
    if (icon->image->mask) {
        XShapeCombineMask(display, icon->iconwindow, ShapeBounding,
	                  0, 0, icon->image->mask, ShapeSet);
    }

    XStoreName(display, icon->iconwindow, "icon pixmap");
    XStoreName(display, icon->textwindow, "icon text");
    XStoreName(display, icon->layer, "icon layer");

    icon->icon_text=NULL; icon->icon_file=NULL; icon->app=NULL;
    icon->line1=NULL; icon->line2=NULL;

    /* Select the event mask for the button */
    XSelectInput(display, icon->iconwindow, PAICONL_EVMASK);
    XSelectInput(display, icon->layer, PAICON_EVMASK);
    XSelectInput(display, icon->textwindow, PAICON_TEXTEVMASK);
    x_addsloot(icon->layer, paicon_handle, icon);
    x_addsloot(icon->iconwindow, paicon_handle, icon);
    x_addsloot(icon->textwindow, paicon_handle, icon);
        
    return icon;
}



void paicon_destroy (PAIcon icon) {

    ASSERT(icon);

    x_delsloot(icon->layer, paicon_handle, icon);
    x_delsloot(icon->textwindow, paicon_handle, icon);
    
    XDestroyWindow(display, icon->iconwindow);
    XDestroyWindow(display, icon->textwindow);
    XDestroyWindow(display, icon->layer);

    x_destroy_pixmap(icon->image);

    if (icon->icon_file) free(icon->icon_file);
    if (icon->icon_text) free(icon->icon_text);
    if (icon->app) free(icon->app);
    if (icon->line1) free(icon->line1);
    if (icon->line2) free(icon->line2);

    memset(icon, '\0', PAIconBytes);
    free(icon);
}


void paicon_setfile(PAIcon icon, const char *name)
{
    ASSERT(icon);
    
    icon->icon_file=malloc(strlen(name)+1);
    if (!icon->icon_file) BUG();

    strcpy(icon->icon_file,name);
}


void paicon_setpos(PAIcon icon, int x, int y)
{
    ASSERT(icon);

    icon->icon_x=(x > (int)DWIDTH ? 100 : x);
    icon->icon_y=(y > (int)DHEIGHT ? 100 : y);
    icon->text_x=icon->icon_x-((TEXTW-icon->image->w)/2);
    icon->text_y=icon->icon_y+icon->image->h+ICON_MARGIN;

    XMoveWindow(display, icon->iconwindow, icon->icon_x, icon->icon_y);
    XMoveWindow(display, icon->layer, icon->icon_x, icon->icon_y);
    XMoveWindow(display, icon->textwindow, icon->text_x, icon->text_y);

}
 

void paicon_seticontext(PAIcon icon, const char *text)
{
    ASSERT(icon);
    
    icon->icon_text=malloc(strlen(text)+1);
    if (!icon->icon_text) die("Not enough memory");
    
    strncpy(icon->icon_text,text,strlen(text)+1);

    /* Split the text in lines */
    paicon_getlines(icon);

}
	
    
void paicon_setapp(PAIcon icon, char **app)
{
    ASSERT(icon);
    ASSERT(app);
    
    icon->app=app;
}


void paicon_mapall(PAIcon icon)
{
    ASSERT(icon);
    
    XMapRaised(display, icon->iconwindow);
    XMapRaised(display, icon->textwindow);
    XMapRaised(display, icon->layer);
}


void paicon_getfile(PAIcon icon, char **name)
{
    ASSERT(icon);
    
    *name=icon->icon_file;
}


void paicon_getpos(PAIcon icon, int *x, int*y)
{
    ASSERT(icon);
    
    *x=icon->icon_x;
    *y=icon->icon_y;

}


static void paicon_exec(PAIcon icon)
{
    pid_t pid;

    ASSERT(icon);
    
    pid=fork();

    if (pid==-1)
	die("fork failed");
    else if (pid==0) {        /* Child */
	/* Make the new fucking X app receive holy signals */
	unblock_signals();

	//fcntl(ConnectionNumber(display), F_SETFD, 1);
        execvp(icon->app[0],icon->app);
	bang("exec of %s failed",icon->app[0]);
	exit(EXIT_FAILURE);
    }
    
}



/* Helpers */

/* Draw the icon pixmap */
static void paicon_draw (PAIcon icon) 
{
    ASSERT(icon);
    
    if (icon->image->pixmap!=None)
        XCopyArea(display, icon->image->pixmap, icon->iconwindow, class.gc,
                  0, 0, icon->image->w, icon->image->h, 0, 0);
}


/* Draw the icon text */
static void paicon_drawtext(PAIcon icon)
{
    
    int line1w, line2w;   /* Text width of both lines after split */
    int line1x, line2x;   /* Position of text */
    
    XClearWindow(display,icon->textwindow);

    /* Center the text */
    line1w=XTextWidth(class.font,icon->line1,(int)strlen(icon->line1));
    line1x=(TEXTW-line1w)/2;
    if (icon->line2) {
        line2w=XTextWidth(class.font,icon->line2,(int)strlen(icon->line2));
        line2x=(TEXTW-line2w)/2;
    }
    
    XDrawString(display,icon->textwindow,class.gc,line1x,
		class.font->ascent+TEXT_TOPMARGIN,
		icon->line1,(int)strlen(icon->line1));		    
    if (icon->line2) 
	XDrawString(display,icon->textwindow,class.gc,line2x,
	            2*class.font->ascent+class.font->descent+TEXT_TOPMARGIN,
		    icon->line2,(int)strlen(icon->line2));		    
	
    
}


static void paicon_getlines (PAIcon icon)
{

    int width;            /* Width of text that will be drawn */
    size_t textlen;       /* Variable just to make code more readable */
    int idx=0, maxidx=0;           
    
    
    ASSERT(icon);
 
    icon->line1=NULL;
    icon->line2=NULL;

    textlen=strlen(icon->icon_text);
    width=XTextWidth(class.font, icon->icon_text,(int) textlen);

    /* Line too long, so we need to cut it a bit ;)) */
    if (width > 2*TEXTI_W) {
        while( XTextWidth(class.font, icon->icon_text,(int)textlen) > (2*TEXTI_W) ) 
 	    textlen--;
	icon->icon_text[textlen]='\0';
	width=XTextWidth(class.font, icon->icon_text, (int)textlen);
    }

    if (width <= TEXTI_W) {   /* Just draw all in one line */
	
        icon->line1=malloc(textlen+1);
        if (!icon->line1) die("Not enough memory");
	
	strncpy(icon->line1,icon->icon_text,textlen);
       	icon->line1[textlen]='\0';
	
    } else {                 /* Split the text in two lines */
	
	while ( XTextWidth(class.font, icon->icon_text, idx) < TEXTI_W )
	    idx++;

	/* Search for blanks in the text and make the break there if 
	   it's possible (i.e lines are not too long)... */
	maxidx=idx;
	idx--;
	while ( (icon->icon_text[idx]!=' ') && idx>=0)
	    idx--;
	if ( XTextWidth(class.font, icon->icon_text+idx+1, strlen(icon->icon_text)-idx) < TEXTI_W )
            maxidx=idx+1;		
	        
        icon->line1=malloc(maxidx+1);
        if (!icon->line1) die("Not enough memory");
        icon->line2=malloc(textlen-maxidx+1);
        if (!icon->line2) die("Not enough memory");
        strncpy(icon->line1,icon->icon_text,(size_t)maxidx); icon->line1[maxidx]='\0';
	strncpy(icon->line2,icon->icon_text+maxidx,textlen-maxidx); icon->line2[textlen-maxidx]='\0';
       
    }
}
