/*
    Filename: pawindow.c
    Synopsis: PA managed Window.

                          Copyright (C) 2000, 2001
            Ral Nez de Arenas Coronado <raul@pleyades.net>
	         David Gmez Espinosa <david@pleyades.net>

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to:

        Free Software Foundation, Inc.
        59 Temple Place, Suite 330
        Boston, MA 02111-1307 USA

*/

#include "pawindow.h"
#include "palabel.h"
#include "pabar.h"
#include "hints.h"


/* Resource definitions: ***************************************************/

/*  - Cursors  */
#define NWCURSOR XC_ul_angle
#define NECURSOR XC_ur_angle
#define SWCURSOR XC_ll_angle
#define SECURSOR XC_lr_angle
#define RESIZECURSOR XC_sizing
#define MOVECURSOR CURSOR

/*  - States */
#define PA_ICONIFIED 	0
#define PA_MAPPED       1
#define PA_WITHDRAWN    2


/****************************************************************************/


/* Minimum border width, in pixels. DO NOT CHANGE */
#define BORDERW (2)

/* These are for easing calculations... */
/* And, don't worry: all pawindow vars are called 'pawindow' XDD */
#define TITLEBAR_X (BORDERW)
#define TITLEBAR_Y (BORDERW)
#define TITLEBAR_W (pawindow->w-BORDERW*2)
#define TITLEBAR_H 20
#define CANVAS_X (BORDERW)
#define CANVAS_Y (BORDERW+TITLEBAR_H)
#define CANVAS_W (pawindow->w-BORDERW*2)
#define CANVAS_H (pawindow->h-BORDERW*2-TITLEBAR_H)
#define SHADOW_X (pawindow->x+BORDERW)
#define SHADOW_Y (pawindow->y+BORDERW)
#define SHADOW_W (pawindow->w)
#define SHADOW_H (pawindow->h)

/* Absolute size limits */
#define FRAME_MIN_W (BORDERW*2+TITLEBAR_H+BUTTONS_W)
#define FRAME_MIN_H (BORDERW*2+TITLEBAR_H+1)
#define FRAME_MAX_W (SHRT_MAX-BORDERW*2)
#define FRAME_MAX_H (SHRT_MAX-BORDERW*2-TITLEBAR_H)

/* Event masks for the pawindow parts */
#define CANVAS_EVMASK (SubstructureRedirectMask | SubstructureNotifyMask| ButtonPressMask | FocusChangeMask)
#define FRAME_EVMASK (ButtonPressMask | ButtonReleaseMask | ExposureMask | EnterWindowMask | LeaveWindowMask)
#define TITLEBAR_EVMASK (ButtonPressMask | ButtonReleaseMask | ExposureMask)


/* Data Structures *********************************************************/

/* This struct contains the per-class data */
static struct {
    XFontStruct *font;
    GC gc;
    struct {
        unsigned long passivebar;
        unsigned long activebar;
        unsigned long passivetext;
        unsigned long activetext;
	unsigned long textshadow;
        unsigned long framelight;
        unsigned long framewhite;
        unsigned long framebase;
        unsigned long framedark;
	unsigned long framedarker;
    } color;
    struct {
        Cursor NW;
        Cursor NE;
        Cursor SW;
        Cursor SE;
        Cursor resize;
	Cursor move;
    } cursor;
    struct {
        xImage close_normal;
        xImage close_active;
        xImage close_pressed;
        xImage iconz_normal;
        xImage iconz_active;
        xImage iconz_pressed;
	xImage maximize_normal;
	xImage maximize_active;
	xImage maximize_pressed;
    } face;  /* Button faces */
    struct {
        Button close;
        Button iconz;
	Button maximize;
    } button;  /* Buttons */
    Bool shadow_status;	          /* shadow status */
    unsigned int titlebar_height; /* Titlebar height */
    unsigned int baseline;        /* Where text is written */
    Time timestamp;               /* Used to check for double-click */
    int clickx, clicky;          
    Chain pawindows;              /* Tracks pawindows */
    Chain fpawindows;             /* Pawindows that have had the focus */
} class;




/* This is the object itself */
struct sPAWindow {
    Window window;       /* Here we put the client window */
    Window shadow;       /* Shadow on the wall... */
    Window frame;        /* Frame around the client */
    Window canvas;       /* Client canvas */
    Window titlebar;     /* Titlebar, of course... */

    Bool active;         /* Has this pawindow the focus? */
    Bool transient;      /* Is this pawindow transient? */
    Bool decorated;      /* Does it need decoration? */
    Bool shaded;         /* Is this pawindow shaded? */
    Bool shaped;         /* Does it have a shape */
    int x, y;            /* Position of a pawindow */
    unsigned int w, h;   /* Dimension of a pawindow */
    unsigned int min_w;  /* Minimum width of client */
    unsigned int min_h;  /* Minimum height of client */
    unsigned int max_w;  /* Maximum width of client */
    unsigned int max_h;  /* Maximum height of client */
    PALabel icon;        /* Iconized form of a pawindow */
    unsigned long state; /* State of a pawindow */
    char *title;         /* Text appearing onto titlebar */
    int oldx, oldy;
    int oldw, oldh;      /* Remember previous geometry , used
			    for un(clefucka)maximizing a pawindow */
    int oldsh;           /* Remember previous height , used 
                            for unshading a pawindow */ 
    
};

#define PAWindowBytes sizeof(struct sPAWindow)

/****************************************************************************/




/* Globals *****************************************************************/

static PAWindow focuspawindow=NULL;
static unsigned short int reference_count=0;

/****************************************************************************/




/* Function prototypes *****************************************************/

/* Handlers: CANVAS */
static void canvas_handle (XEvent *, void *);
static void canvas_DestroyNotify    (XDestroyWindowEvent *, PAWindow);
static void canvas_UnmapNotify      (XUnmapEvent *, PAWindow);
static void canvas_ConfigureRequest (XConfigureRequestEvent *, PAWindow);
static void canvas_PropertyNotify   (XPropertyEvent *, PAWindow);
static void canvas_MapRequest       (XMapEvent *, PAWindow);
static void canvas_ClientMessage    (XClientMessageEvent *, PAWindow);
static void canvas_FocusOut         (XFocusOutEvent *, PAWindow);

/* Handlers: FRAME */
static void frame_handle (XEvent *, void *);
static void frame_ButtonPress (XButtonPressedEvent *, PAWindow);
static void frame_Expose      (XExposeEvent *, PAWindow);
static void frame_EnterNotify (XEnterWindowEvent *, PAWindow);
static void frame_LeaveNotify (XEnterWindowEvent *, PAWindow);

/* Handlers: TITLEBAR */
static void titlebar_handle (XEvent *, void *);
static void titlebar_ButtonPress (XButtonPressedEvent *, PAWindow);
static void titlebar_Expose      (XExposeEvent *, PAWindow);



/* Drawing functions */
static void draw_titlebar (PAWindow);
static void draw_border   (PAWindow);
static void draw_outline  (int, int, int, int);

/* Resize and move helpers */
static void take_a_trip        (PAWindow);
static void stretch_and_shrink (PAWindow);

/* Other helpers */
static void kill_client (void *);
static void iconify_client (void *);
static void maximize_client (void *);
static void restore_client (void *);
static void set_title (PAWindow, XTextProperty *);
static void set_shape (PAWindow);

/****************************************************************************/


/* Inline functions for using the chain with stack semantics */
inline void chain_push(Chain chain, void *object)
{
    PAWindow pw;
    
    ASSERT(chain);
    ASSERT(object);

    pw= (PAWindow) object;

    chain_home(chain);
    chain_put(chain, pw);
}


inline void *chain_pop(Chain chain)
{
    PAWindow pw;
    
    ASSERT(chain);

    chain_seek(chain,1);
    pw=chain_del(chain);

    return pw;
}


/****************************************************************************/
/*                                                                          */
/*                        Event handling routines                          */
/*                                                                          */
/****************************************************************************/

/* Canvas event handling ***************************************************/
static void canvas_handle (XEvent *xevent, void *widget) {

    PAWindow pawindow=NULL;
    int type, foo;

    ASSERT(xevent);
    ASSERT(widget);

    pawindow=(PAWindow)widget;

    /* It seems that handling shape events its really not needed to support
       shaped windows... Just check for out own generated Shape events */
    XShapeQueryExtension(display,&type,&foo);
    if (xevent->type==type) {
	GOT(ShapeNotify,Canvas);
	return;
    }
    
    switch (xevent->type) {
    case DestroyNotify:
        GOT(DestroyNotify, Canvas);
        canvas_DestroyNotify(&(xevent->xdestroywindow), pawindow);
        break;
    case UnmapNotify:
        GOT(UnmapNotify, Canvas);
        canvas_UnmapNotify(&(xevent->xunmap), pawindow);
        break;
    case ConfigureRequest:
        GOT(ConfigureRequest, Canvas);
        canvas_ConfigureRequest(&(xevent->xconfigurerequest), pawindow);
        break;
    case PropertyNotify:
        GOT(PropertyNotify, Canvas);
        canvas_PropertyNotify(&(xevent->xproperty), pawindow);
        break;
    case ButtonPress:
	GOT(ButtonPress, Canvas); /* Just ignore it to avoid propagate to frame*/	
	break;
    case MapRequest:
	GOT(MapRequest, Canvas);
	canvas_MapRequest(&(xevent->xmap), pawindow);
	break;
    case ClientMessage:
	GOT(ClientMessage, Canvas);
        canvas_ClientMessage(&(xevent->xclient), pawindow);
	break;
    case FocusIn:
	GOT(FocusIn, Canvas);
	break;
    case FocusOut:
	GOT(FocusOut, Canvas);
	canvas_FocusOut(&(xevent->xfocus), pawindow);
	break;
    default:
	SAY("Event type %d not handled on Canvas",xevent->type);
	break;
    }
}



static void canvas_DestroyNotify (XDestroyWindowEvent *event,
                                  PAWindow pawindow) {

    ASSERT(event);
    ASSERT(pawindow);    
    if (event->window!=pawindow->window) BUG();

    pawindow_destroy(pawindow);

}




static void canvas_UnmapNotify (XUnmapEvent *event, PAWindow pawindow) {

    XEvent xevent;

    ASSERT(event);
    ASSERT(pawindow);
    if (event->window!=pawindow->window) BUG();
     
    /* Ignore UnmapNotify coming from pawindows being shaded, but
       keeping and eye on the window being destroyed */
    if (pawindow->shaded) return;

    pawindow_unmap(pawindow);

    /* See if the client wants to go Withdrawn */
    /*
        As per ICCCM, when a ICCCM-aware client wants to withdraw one
    of its windows it should unmap the window and, after that, send a
    synthetic UnmapNotify to the root window. This is quite congruent,
    and the proper way of dealing with this protocol is to handle the
    event standing in one leg, with the arms on the back, telling three
    times 'Klaatu verata nictu" (reversed), casting a fireball spell on
    the XServer and invoking the Ancient Ones while you do the XRequest...
    */
    if (XCheckTypedWindowEvent(display, ROOT, UnmapNotify, &xevent)==True) {
 
      CARD32 data[2];  

      SAY("Going withdrawn");
      pawindow->state=PA_WITHDRAWN;
      pabar_dellabel(pawindow->icon);
      data[0]=WithdrawnState;
      data[1]=None;
      XChangeProperty(display, pawindow->window, wm_state, wm_state,
                      32, PropModeReplace, (unsigned char *)data, 2);
    }

}



static void canvas_ConfigureRequest (XConfigureRequestEvent *event,
                                     PAWindow pawindow) {

    int x=0, y=0;
    unsigned int w=0, h=0;

    ASSERT(event);
    ASSERT(pawindow);
    if (event->window!=pawindow->window) BUG();

    /* Use relative coordinates as default values */
    x=(pawindow->decorated?pawindow->x+BORDERW:pawindow->x);
    y=(pawindow->decorated?pawindow->y+BORDERW+TITLEBAR_H:pawindow->y);
    w=(pawindow->decorated?pawindow->w-BORDERW*2:pawindow->w);
    h=(pawindow->decorated?pawindow->h-BORDERW*2-TITLEBAR_H:pawindow->h);
        
    /* Get only useful values (specified in the value_mask) */
    if (event->value_mask & CWX || event->value_mask & CWY) {
       	x=event->x; 
	y=event->y;
	pawindow_move(pawindow, x, y);
    }
    
    /* Make sure the client fits on the display when it's configuring itself */
    if (event->value_mask & CWWidth || event->value_mask & CWHeight) {
	SAY("Requested width: %d",event->width);
        if ( ((unsigned int)event->width+BORDERW*2) > DWIDTH ) {
	    w=DWIDTH-BORDERW*2;
	    x=BORDERW;
	}
        else	
	    w=event->width;
	
	SAY("Requested height: %d",event->height);
        if ( ((unsigned int)event->height+BORDERW*2+TITLEBAR_H) > DHEIGHT ) {
	    h=DHEIGHT-BORDERW*2-TITLEBAR_H;
	    y=0;
	}
        else	
	    h=event->height;

	SAY("Resizing to w: %d h: %d",w,h);
	pawindow_resize(pawindow, w, h);
    } else {

        /* Some windows in withdrawn state generate a ConfigureRequest event,
           but no MapRequest, so it's necessary to map them or they'll never
           become visible. Welcome to The X Horror Show */
        XGrabServer(display);
        pawindow_map(pawindow);
        XUngrabServer(display);
    }
    
    XFlush(display);
    
}




static void canvas_PropertyNotify (XPropertyEvent *event, PAWindow pawindow) {

    char *atomname;
	
    ASSERT(event);
    ASSERT(pawindow);
    if (event->window!=pawindow->window) BUG();

    atomname=XGetAtomName(display,event->atom);
    SAY("Atom received: %s",atomname);

    switch (event->atom) {
    case XA_WM_TRANSIENT_FOR: {  /* Check for transient windows */
        Window w=None;
        if (XGetTransientForHint(display, event->window, &w)) {
            pawindow->transient=(w==None) ? False : True;
            draw_titlebar(pawindow);  /* Different decoration */
        }
	break;
    }
    case XA_WM_NAME: {  /* Get a new name for the window */
        XTextProperty name;

        if (!XGetWMName(display, event->window, &name)) return;
        set_title(pawindow, &name);
	palabel_settext(pawindow->icon, (char *)(name.value), name.nitems);
        if (name.value) XFree(name.value);
        break;
    }
    case XA_WM_ICON_NAME: {  /* New name for the icon */
        XTextProperty name;
       
        if (!XGetWMIconName(display, event->window, &name)) return;
        palabel_settext(pawindow->icon, (char *)(name.value), name.nitems);
        if (name.value) XFree(name.value);
        break;
    }
    case XA_WM_NORMAL_HINTS: {  /* Get the size hints for the window */
        XSizeHints *hints=NULL;
        long mask=0;

        hints=XAllocSizeHints();
        if (!hints) return;
        if (!XGetWMNormalHints(display, event->window, hints, &mask)) return;

        /* Get size limits. Ignoring mask value and using instead flags 
	 * defined in XSizeHints */
	if (hints->flags & PMinSize) {
            pawindow->min_w=hints->min_width;
            pawindow->min_h=hints->min_height;

	    /* Only sanitize values when the pawindow is decorated, cause
	     * non-decorated windows are allowed to be as small as they want */
            if (pawindow->decorated) {
             if (pawindow->min_w < FRAME_MIN_W) pawindow->min_w=FRAME_MIN_W;
             if (pawindow->min_h < FRAME_MIN_H) pawindow->min_h=FRAME_MIN_H;
             if (pawindow->min_w > FRAME_MAX_W) pawindow->min_w=FRAME_MAX_W;
             if (pawindow->min_h > FRAME_MAX_H) pawindow->min_h=FRAME_MAX_H;
	    }
	    
        }

        if (hints->flags & PMaxSize) {
            pawindow->max_w=hints->max_width;
            pawindow->max_h=hints->max_height;

            if (pawindow->max_w==0) pawindow->max_w=FRAME_MAX_W;
            if (pawindow->max_h==0) pawindow->max_h=FRAME_MAX_H;
            if (pawindow->max_w < FRAME_MIN_W) pawindow->max_w=FRAME_MIN_W;
            if (pawindow->max_h < FRAME_MIN_H) pawindow->max_h=FRAME_MIN_H;
            if (pawindow->max_w > FRAME_MAX_W) pawindow->max_w=FRAME_MAX_W;
            if (pawindow->max_h > FRAME_MAX_H) pawindow->max_h=FRAME_MAX_H;
        }

        /* Adjust values. Some clients have a bad attitude. */
        if (pawindow->max_w < pawindow->min_w)
            pawindow->max_w=pawindow->min_w;
        if (pawindow->max_h < pawindow->min_h)
            pawindow->max_h=pawindow->min_h;

        XFree(hints);
	break;
    }
    }
}


static void canvas_MapRequest(XMapEvent *event, PAWindow pawindow)
{
	
    ASSERT(event);
    ASSERT(pawindow);
    if (event->window!=pawindow->window) BUG();

    /* First time a window is created, MapRequest arrives at root window,
     * and any other MapRequest will be send to the canvas (cause is his parent 
     * window and has SubstructureRedirectMask selected) */

    XGrabServer(display);
    pawindow_map(pawindow);
    XUngrabServer(display);
    
}


static void canvas_ClientMessage(XClientMessageEvent *event, PAWindow pawindow)
{

    ASSERT(event);
    ASSERT(pawindow);
    if (event->window!=pawindow->window) BUG();

    if (event->data.l[0]==IconicState)
	pawindow_unmap(pawindow);

}


static void canvas_FocusOut(XFocusOutEvent *event, PAWindow pawindow)
{
    
    ASSERT(event);
    ASSERT(pawindow);
    
    /* Some toolkits like the fucking XITK steal the focus, some we have
       to take it back again to pawm */
    if (event->detail==NotifyNonlinear) {
        if (event->window!=pawindow->window) {SAY("focus out not in pawindow");}
        if (focuspawindow) {
	    SAY("taking back focus");
            XSetInputFocus(display,focuspawindow->window,
	  	           RevertToPointerRoot,CurrentTime); }
    }
}



/****************************************************************************/


/* Frame event handling ****************************************************/
static void frame_handle (XEvent *xevent, void *widget) {

    PAWindow pawindow=(PAWindow)widget;
    
    ASSERT(xevent);
    ASSERT(pawindow);

    switch (xevent->type) {
    case ButtonPress:
        GOT(ButtonPress, Frame);
        frame_ButtonPress(&(xevent->xbutton), pawindow);
        break;
    case Expose:
        GOT(Expose, Frame);
        frame_Expose(&(xevent->xexpose), pawindow);
        break;
    case EnterNotify:
	GOT(EnterNotify, Frame);
	frame_EnterNotify(&(xevent->xcrossing), pawindow);
        break;
    case LeaveNotify:
	GOT(LeaveNotify, Frame);
	frame_LeaveNotify(&(xevent->xcrossing), pawindow);
        break;
    default:
	SAY("Event %d on Frame",xevent->type);
        break;	
 
    }
}




static void frame_ButtonPress (XButtonPressedEvent *event, PAWindow pawindow)
{
    ASSERT(event);
    ASSERT(pawindow);    

    if (event->button!=Button1) return;
    if (event->window!=pawindow->frame) BUG();

    /* A click on the frame border causes a resize, but a click on
       the window (frame child) causes a click-to focus policy */
    if ( ((event->x < BORDERW) ||                           // Left border   
	 (event->y < BORDERW) ||                            // Top border    
	 (event->x >= (int)(pawindow->w-BORDERW)) ||        // Right border  
	 (event->y >= (int)(pawindow->h-BORDERW)) ) &&   // Bottom border 
	 pawindow->decorated) {               
	
        XGrabServer(display);
        pawindow_map(pawindow);
        XUngrabServer(display);

        /* Resize the pawindow */
        stretch_and_shrink(pawindow);
	/* And make sure that the pointer is ungrabbed with shaded windows */
	if (pawindow->shaded) 
	     XUngrabPointer(display, CurrentTime);
    } else {

	/* Click-to-focus :)) */
        //XGrabServer(display);
        pawindow_map(pawindow);
        //XUngrabServer(display);

	/* This must be done in the last place, or bad things can
	   happen (i.e. ButtonRelase events are not reported to the
	   window where the ButtonPress was delivered) */
        XAllowEvents(display, ReplayPointer, CurrentTime);
    }

}




static void frame_Expose (XExposeEvent *event, PAWindow pawindow) 
{
    ASSERT(event);
    ASSERT(pawindow);

    /* Ugly return for decorated windows, could it be avoided?*/
    if (event->window!=pawindow->frame) return;
        
    /* More Expose events to come */
    //if (event->count) return;

    /* Update the frame art ;) */
    draw_border(pawindow);
}



static void frame_EnterNotify (XEnterWindowEvent *event, PAWindow pawindow)
{
    ASSERT(event);
    ASSERT(pawindow);

    /* Ditto, see frame_Expose... */
    if (event->window!=pawindow->frame) return;

    XGrabButton(display, Button1, AnyModifier, pawindow->frame, False,
	        ButtonPressMask, GrabModeSync, GrabModeSync, None, None);
}

static void frame_LeaveNotify(XEnterWindowEvent *event, PAWindow pawindow)
{
    ASSERT(event);
    ASSERT(pawindow);

    /* Sigh... X S-U-C-K-S */
    if (event->window!=pawindow->frame) return;

    /* Ungrab? No thanks XDD. The passive grab created by XGrabButton
       changes to an active grab, which is automagically released when
       a ButtonRelease event is received */
}
    

/****************************************************************************/



/* Titlebar event handling *************************************************/
static void titlebar_handle (XEvent *xevent, void *widget) {

    PAWindow pawindow=NULL;

    ASSERT(xevent);
    ASSERT(widget);

    pawindow=(PAWindow)widget;

    switch (xevent->type) {
    case ButtonPress:
        GOT(ButtonPress, Titlebar);
        titlebar_ButtonPress(&(xevent->xbutton), pawindow);
        break;
    case Expose:
        GOT(Expose, Titlebar);
        titlebar_Expose(&(xevent->xexpose), pawindow);
        break;
    default:
	SAY("Event %d on Titlebar",xevent->type);
        break;	
    }
}




static void titlebar_ButtonPress (XButtonPressedEvent *event, PAWindow pawindow) 
{
    ASSERT(event);
    ASSERT(pawindow);

    if (event->window!=pawindow->titlebar) BUG();
    if (event->button!=Button1) return;

    XGrabServer(display);
    pawindow_map(pawindow);
    XUngrabServer(display);

    /* These thresholds should probably be configurable */
    if ( (labs(event->time-class.timestamp) < 400) &&
	 (abs(event->x_root-class.clickx) < 5) && 
	 (abs(event->y_root-class.clicky) < 5) ) 
    {
        if (pawindow->shaded) 
	    pawindow_unshade(pawindow); /* Unshade the window */
        else 
	    pawindow_shade(pawindow);   /* Shade this fucking window */
	class.timestamp=0;
	class.clickx=-1;
	class.clicky=-1;
	
    } else {
	class.timestamp=event->time;
	class.clickx=event->x_root;
	class.clicky=event->y_root;
        /* Make a fucking travel :))) */
        take_a_trip(pawindow);
    }
}




static void titlebar_Expose (XExposeEvent *event, PAWindow pawindow) {

    ASSERT(event);
    ASSERT(pawindow);
    if (event->window!=pawindow->titlebar) BUG();

    /* More Expose events to come. */
    if (event->count) return;

    /* Update the titlebar art ;) */
    draw_titlebar(pawindow);

}
/****************************************************************************/



/****************************************************************************/
/*                                                                          */
/*                      Private (static) functions                         */
/*                                                                          */
/****************************************************************************/



/* Drawing functions *******************************************************/

/* NOTE: ANY of the draw_*() functions can modify the GC. Be careful... */

/* This one redraws the title, if any */
static void draw_titlebar (PAWindow pawindow) 
{
    unsigned int w=TITLEBAR_W;  /* Maximum width available for title */
    int x=0;
    size_t len=0;

    ASSERT(pawindow);
    
    if (!pawindow->decorated) return;

    /* Draw politely */
    XSetFunction(display, class.gc, GXcopy);
    XSetSubwindowMode(display, class.gc, ClipByChildren);

    /* Active windows have different decorations */
    if (pawindow->active==True) XSetWindowBackground(display,
                                                     pawindow->titlebar,
                                                     class.color.activebar);
    else XSetWindowBackground(display,
                              pawindow->titlebar,
                              class.color.passivebar);

    XClearWindow(display, pawindow->titlebar);
    if (!(pawindow->title)) return;

    /* Compute text length allowed for avoiding cut-text */
    /* First: non-transient active windows have buttons, so we have
       least space, the rest of windows have avaliable the full width */
    if (( !pawindow->transient && pawindow->active ))  
        w-=BUTTONS_W;
    /* Second: we want a little margin around text */
    w-=2*(class.font->descent);  /* Leave a pretty margin */

    len=x_trimtext(class.font, pawindow->title, w);

    /* Now center the title */
    /* 'w' contains the available space, and 'len' the text length */
    x=(int) w-XTextWidth(class.font, pawindow->title, (int) len);
    x/=2;
    
    /* Special decorations for active/passive windows */
    if (pawindow->active==True) {
        /* The text has a shadow on active windows */
        XSetForeground(display, class.gc, class.color.textshadow);
        XDrawString(display, pawindow->titlebar, class.gc,
                    x+2, (int)class.baseline+2,
                    pawindow->title,(int)len);
        XSetForeground(display, class.gc, class.color.activetext);
    } else XSetForeground(display, class.gc, class.color.passivetext);

    /* Finally add the text */
    XDrawString(display, pawindow->titlebar, class.gc,
                x, (int)class.baseline, pawindow->title,(int)len);
}




/* Draw the Puto Amo(tm) 3D border around the Puto Amo(tm) window */
static void draw_border (PAWindow pawindow) {

    unsigned int w=pawindow->w;
    unsigned int h=pawindow->h;

    if (!pawindow->decorated) return;

    XSetFunction(display, class.gc, GXcopy);

    /*
        NOTE: the amazingly stupid X Window Fucking System has a very
    useful feature: If you paint a line from left to right the last point
    is not painted (is an X convention, I suppose), but if you paint the
    very same line from right to left, the first point misses :))))

        This way, you end up with absurd endpoint specs and being *very*
    careful about how you draw a line :) The X Window System is quite useful,
    no doubt...
    */

    /* NorthWest */
    /* 0 */
    XSetForeground(display, class.gc, class.color.framelight);
    XDrawLine(display, pawindow->frame, class.gc, 0, 0,(int) w, 0);
    XDrawLine(display, pawindow->frame, class.gc, 0, 0, 0,(int) h);

    /* 1 */
    XSetForeground(display, class.gc, class.color.framedark);
    XDrawLine(display, pawindow->frame, class.gc, 1, 1,(int) w-1, 1);
    XDrawLine(display, pawindow->frame, class.gc, 1, 1, 1,(int) h-1);

    /* SouthEast */
    /* 0 */
    XSetForeground(display, class.gc, class.color.framedarker);
    XDrawLine(display, pawindow->frame, class.gc, 0,(int) h-1,(int)w,(int)h-1);
    XDrawLine(display, pawindow->frame, class.gc,(int) w-1, 0,(int)w-1,(int)h);

    /* 1 */ /* 2 */
    XSetForeground(display, class.gc, class.color.framewhite);
    XDrawLine(display, pawindow->frame, class.gc, 1,(int) h-2,(int)w-2,(int)h-2);
    XDrawLine(display, pawindow->frame, class.gc,(int) w-2, 1,(int)w-2,(int)h-2);

}




static void draw_outline (int x1, int y1, int x2, int y2) {

    unsigned int w=(x2-x1);
    unsigned int h=(y2-y1);

    XSetSubwindowMode(display, class.gc, IncludeInferiors);
    XSetFunction(display, class.gc, GXinvert);
    XDrawRectangle(display, ROOT, class.gc, x1, y1, w, h);
    x1+=BORDERW-1;
    y1+=BORDERW-1;
    w-=(BORDERW-1)*2;
    h-=(BORDERW-1)*2;
    XDrawRectangle(display, ROOT, class.gc, x1, y1, w, h);

}
/****************************************************************************/




/* Geometry helper functions ***********************************************/
static void take_a_trip (PAWindow pawindow) {

    int x1=0, y1=0, x2=0, y2=0;
    int xoff=0, yoff=0;
    long evmask;
    unsigned int tmpmask;
    Window root=None;
    Window window=None;
    Cursor cursor=class.cursor.move;

    ASSERT(pawindow);
    
    /* Unmap the buttons... */
    button_unmap(class.button.close);
    button_unmap(class.button.iconz);
    button_unmap(class.button.maximize);

    /* Reset the saved geometry */
    pawindow->oldx=0; pawindow->oldy=0;
    pawindow->oldw=0; pawindow->oldh=0;

    /* Where the fuck is the fucking pointer???? */
    if (XQueryPointer(display, pawindow->frame, &root, &window,
                      &x1, &y1, &x2, &y2, &tmpmask)==False) BUG();

    if ((x1 < 0) || (y1 < 0)) BUG();

    /* Determine pointer offset */
    xoff=x1-pawindow->x;
    yoff=y1-pawindow->y;

    /* Determine corners for the outline */
    x1=pawindow->x;
    y1=pawindow->y;
    x2=pawindow->x+pawindow->w-1;
    y2=pawindow->y+pawindow->h-1;

    draw_outline(x1, y1, x2, y2);

    /* Special event mask for doing the move */
    evmask=ButtonPressMask|ButtonReleaseMask|PointerMotionMask;

    /* Capture both pointer and XServer . By the way, X data types are
     * a huge pile of shit. Event mask is a long in X*Event functions, but 
     * in XGrabPointer is an unsigned int */
    XGrabPointer(display, ROOT, True, (unsigned int)evmask,
                 GrabModeAsync, GrabModeAsync, None,
                 cursor, CurrentTime);
    XGrabServer(display);

    while (1) {  /* Loops 'til button released */
        XEvent xevent;
        static Bool moved=False;

        XMaskEvent(display, evmask, &xevent);

        switch (xevent.type) {

        case MotionNotify: {  /* If pointer moved, move outline */
            XMotionEvent event=xevent.xmotion;

            moved=True;
            draw_outline(x1, y1, x2, y2);

            x1=event.x_root-xoff;
            y1=event.y_root-yoff;
            x2=x1+pawindow->w-1;
            y2=y1+pawindow->h-1;

            draw_outline(x1, y1, x2, y2);
            break;
        }

        case ButtonRelease: {  /* End the moving */
				
	    XButtonPressedEvent *event=&(xevent.xbutton);
	   
	    if (event->button!=Button1) break;

            draw_outline(x1, y1, x2, y2);

            /* Translate to client coordinates */
	    if (pawindow->decorated) {
                x1+=BORDERW;
                y1+=BORDERW+TITLEBAR_H;
	    }

            if (moved==True) pawindow_move(pawindow, x1, y1);

            button_map(class.button.close);
            button_map(class.button.iconz);
	    button_map(class.button.maximize);

	    XUngrabServer(display);
            XUngrabPointer(display, CurrentTime);
            return;
        }
        }
    }
}




/* Resize the pawindow */
static void stretch_and_shrink (PAWindow pawindow) {

    /* Well, gravity is not needed at all */
    Cursor cursor=None;
    int x=0;
    int y=0;
    int x1=0, y1=0, x2=0, y2=0;  /* Outline corners */
    /* Variable aliases to improve readability... */
    int min_w=pawindow->min_w+BORDERW*2;
    int min_h=pawindow->min_h+BORDERW*2+TITLEBAR_H;
    int max_w=pawindow->max_w+BORDERW*2;
    int max_h=pawindow->max_h+BORDERW*2+TITLEBAR_H;
    Window root=None;
    Window window=None;
    unsigned int tmpmask;
    long evmask;

    ASSERT(pawindow);

    /* Avoid resizing if the pawindow is shaded */
    if (pawindow->shaded) return;

    /* Unmap the buttons... */
    button_unmap(class.button.close);
    button_unmap(class.button.iconz);
    button_unmap(class.button.maximize);

    /* Find the pointer */
    if (XQueryPointer(display, pawindow->frame, &root, &window,
                      &x1, &y1, &x, &y, &tmpmask)==False) BUG();

    if (x < 0) x=0;
    if (y < 0) y=0;

    /* Find the corner to warp pointer to */
    x=(x > (int)pawindow->w/2) ? pawindow->w-1 : 0;
    y=(y > (int)pawindow->h/2) ? pawindow->h-1 : 0;

    /* Put a pretty cursor depending on the corner ;) */
    cursor=(x==0) ? ((y==0) ? class.cursor.NW : class.cursor.SW):
                    ((y==0) ? class.cursor.NE : class.cursor.SE);

    /* Move the pointer to the nearest corner */
    XWarpPointer(display, pawindow->frame, pawindow->frame,
                 0, 0, pawindow->w, pawindow->h, x, y);

    /* Nice event mask to manage mouse temporarily */
    evmask=ButtonPressMask|ButtonReleaseMask|PointerMotionMask;

    /* Do the grabbings */
    XGrabPointer(display, ROOT, True, (unsigned int)evmask,
                 GrabModeAsync, GrabModeAsync, None,
                 cursor, CurrentTime);
    XGrabServer(display);

    /* Calculate corner coordinates */
    x1=pawindow->x;
    y1=pawindow->y;
    x2=pawindow->w+x1-1;
    y2=pawindow->h+y1-1;
    draw_outline(x1, y1, x2, y2);

    while (1) {  /* Loop 'til button released */
        XEvent xevent;
        XMaskEvent(display, evmask, &xevent);

        switch (xevent.type) {
        case MotionNotify: {
            XMotionEvent event=xevent.xmotion;

            draw_outline(x1, y1, x2, y2);

            x=event.x;
            y=event.y;

            /* Calculate outline on NorthWest corner */
            if (cursor==class.cursor.NW) {
                x1=x;
                y1=y;

                if (x2-x1 > max_w-1) x1=x2-max_w+1;
                if (x2-x1 < min_w-1) x1=x2-min_w+1;
                if (y2-y1 > max_h-1) y1=y2-max_h+1;
                if (y2-y1 < min_h-1) y1=y2-min_h+1;
            }
                        
            /* Calculate outline on NorthEast corner */
            if (cursor==class.cursor.NE) {
                x2=x;
                y1=y;
                                
                if (x2-x1 > max_w-1) x2=x1+max_w-1;
                if (x2-x1 < min_w-1) x2=x1+min_w-1;
                if (y2-y1 > max_h-1) y1=y2-max_h+1;
                if (y2-y1 < min_h-1) y1=y2-min_h+1;
            }
                        
            /* Calculate outline on SouthWest corner */
            if (cursor==class.cursor.SW) {
                x1=x;
                y2=y;
                                
                if (x2-x1 > max_w-1) x1=x2-max_w+1;
                if (x2-x1 < min_w-1) x1=x2-min_w+1;
                if (y2-y1 > max_h-1) y2=y1+max_h-1;
                if (y2-y1 < min_h-1) y2=y1+min_h-1;
            }
                        
            /* Calculate outline on SouthEast corner */
            if (cursor==class.cursor.SE) {

                x2=x;
                y2=y;
                               
                if (x2-x1 > max_w-1) x2=x1+max_w-1;
                if (x2-x1 < min_w-1) x2=x1+min_w-1;
                if (y2-y1 > max_h-1) y2=y1+max_h-1;
                if (y2-y1 < min_h-1) y2=y1+min_h-1;
            }

            draw_outline(x1, y1, x2, y2);
            break;
        }

        case ButtonRelease: {
            XButtonPressedEvent *event=&(xevent.xbutton);
                        
            if (event->button!=Button1) break;
            draw_outline(x1, y1, x2, y2);

            /* Translate to client coordinates */
	    if (pawindow->decorated) {
                x1+=BORDERW;
                y1+=BORDERW+TITLEBAR_H;
                x2-=BORDERW;
                y2-=BORDERW;
	    }

	    /* Reset saved position before resizing */
	    pawindow->oldx=0; pawindow->oldy=0;
	    pawindow->oldw=0; pawindow->oldh=0;

            pawindow_move(pawindow, x1, y1);
            pawindow_resize(pawindow, (unsigned int)x2-x1+1,(unsigned int)y2-y1+1);
            pawindow_map(pawindow);
	    XUngrabServer(display);
            XUngrabPointer(display, CurrentTime);
            return;
        }
        }
    }
}
/****************************************************************************/


/* Other helpers ***********************************************************/
/* This one kills a client */
static void kill_client (void *pw) {

    PAWindow pawindow=(PAWindow)pw;
    Atom *proto=NULL;
    int n=0;
        
    ASSERT(pw);

    /* See whether the client accepts WM protocols */        
    if (XGetWMProtocols(display, pawindow->window, &proto, &n)) {
        int i=0;
        for (i=0; i<n; i++) {
            if (proto[i]==wm_delete) {
                XEvent e;

                /* Build and send the WM_DELETE command to the client */
		SAY("WM_DELETE sent");
                e.type=ClientMessage;
                e.xclient.window=pawindow->window;
                e.xclient.message_type=wm_protocols;
                e.xclient.format=32;
                e.xclient.data.l[0]=wm_delete;
                e.xclient.data.l[1]=CurrentTime;

                if (!XSendEvent(display, pawindow->window, False,
                                NoEventMask, &e)) BUG();
                /* Shaded means we'll not get any damn UnmapNotify so...*/
                if (pawindow->shaded)
		    pawindow_unmap(pawindow);
                break;
            }
            /* Atom not found, just kill */
            if (i==n) XKillClient(display, pawindow->window);
        }
    } else XKillClient(display, pawindow->window);  /* Fuck out */
}


static void maximize_client (void *pw) {

    PAWindow pawindow=(PAWindow)pw;
    char *tmpvalue;

    ASSERT(pw);

    if (pawindow->shaded) 
	pawindow_unshade(pawindow);
    
    else if (!pawindow->oldw) {
        /* Keep previous geometry */
        pawindow->oldx=pawindow->x;
        pawindow->oldy=pawindow->y;
        pawindow->oldw=pawindow->w;
        pawindow->oldh=pawindow->h;
	
        /* Pawindow functions use client relative coordinates... */
        pawindow_move(pawindow, BORDERW, BORDERW-TITLEBAR_H );
	/* Don't obscure the pabar when maximizing */
	tmpvalue=cfgpool_getstring(pawmpool, "pabar_autohide");
	if (strcmp(tmpvalue,"On")==0) 
            pawindow_resize(pawindow, DWIDTH-(BORDERW*2),
	   	                      LOWBAR-(BORDERW*2)-TITLEBAR_H);
	else
            pawindow_resize(pawindow, DWIDTH-(BORDERW*2),
	   	                      TOPBAR-(BORDERW*3)-TITLEBAR_H);
	free(tmpvalue);
    }
    else {        /* Restore previous geometry */
        pawindow->x=pawindow->oldx; pawindow->oldx=0;
        pawindow->y=pawindow->oldy; pawindow->oldy=0;
        pawindow->w=pawindow->oldw; pawindow->oldw=0;
        pawindow->h=pawindow->oldh; pawindow->oldh=0;

	pawindow_move(pawindow, pawindow->x+BORDERW, pawindow->y+BORDERW+TITLEBAR_H);
	pawindow_resize(pawindow, pawindow->w-(BORDERW*2),
				  pawindow->h-(BORDERW*2)-TITLEBAR_H);
    }
 
}



static void iconify_client (void *pw) {

    PAWindow pawindow=(PAWindow)pw;
    
    ASSERT(pw);
    
    pawindow_unmap(pawindow);

}




static void restore_client (void *pw) {

    PAWindow pawindow=(PAWindow)pw;
    
    ASSERT(pw);
    
    if ((pawindow->state == PA_MAPPED) && pawindow->active) 
        pawindow_unmap(pawindow);
    else 
        pawindow_map(pawindow);
    
}




/* This function sets up a new title onto the title bar */
static void set_title (PAWindow pawindow, XTextProperty *text) {

    ASSERT(pawindow);
	
    if (pawindow->title) free(pawindow->title);
    pawindow->title=NULL;
        
    /* Assert some values */
    if (!text) return;
    if (!(text->value)) return;
    if (text->encoding != XA_STRING) return;
    if (text->format != 8) return;

    /* Copy the string */
    pawindow->title=malloc(1+text->nitems);
    if (!(pawindow->title)) die("Not enough memory");  /* Huy que mal rollito... */

    strncpy(pawindow->title, text->value, text->nitems);
    pawindow->title[text->nitems]='\0';
    draw_titlebar(pawindow);

}


/* Give the keyboard focus to the PAwindow passed as parameter */
static void change_focus(PAWindow pawindow) 
{

    ASSERT(pawindow);

    /* The pawindow already has the focus */
    if (focuspawindow && focuspawindow==pawindow) {
	if (pawindow->shaded)
            XSetInputFocus(display, pawindow->frame, RevertToPointerRoot, CurrentTime);
	else
            XSetInputFocus(display, pawindow->window, RevertToParent, CurrentTime);
	return;
    }
    
    if (focuspawindow) {
        focuspawindow->active=False;
	palabel_unsetfocus(focuspawindow->icon);
        /* A BadWindow could happen here (could==always in X Window notation),
         * when the unmapped window has been already destroyed by the client. 
         * X kicking asses ;) in the good olde tradition */
        draw_titlebar(focuspawindow);
    }
    
    focuspawindow=pawindow;
    SAY("Pawindow %#lx pushed",pawindow->window);
    chain_push(class.fpawindows, pawindow);

    if (pawindow->shaded) 
        XSetInputFocus(display, pawindow->frame, RevertToPointerRoot, CurrentTime);
    else 
        XSetInputFocus(display, pawindow->window, RevertToParent, CurrentTime);

    palabel_setfocus(pawindow->icon);

}


/* Configures the shape of a pawindow, if needed */
static void set_shape(PAWindow pawindow)
{
    XRectangle *shp_rectangles;    /* Shape information */
    int num,order;
    int major,minor;

    ASSERT(pawindow);

    XShapeQueryVersion(display, &major, &minor);
    SAY("Configuring shape... (XShape version %d.%d)",major,minor);

    /* First we get the shape of the client window, and after use it to set
       the shape of the pawindow canvas and frame. We don't do a ShapeClip
       because X window borders are not being used by pawm */
    SAY("Setting bounding shape in pawindow");
    shp_rectangles=XShapeGetRectangles(display, pawindow->window, 
                                       ShapeBounding, &num, &order);
    if (shp_rectangles==NULL) return;
    SAY("set_shape: shaping canvas...");
    XShapeCombineRectangles(display, pawindow->canvas, ShapeBounding,
                            0, 0, shp_rectangles, num, ShapeSet, order);
    SAY("set_shape: shaping frame...");
    XShapeCombineRectangles(display, pawindow->frame, ShapeBounding,
        	            0, 0, shp_rectangles, num, ShapeSet, order);
    SAY("set_shape: shaping done");
    XFree(shp_rectangles);
    SAY("set_shape: exiting...");

}

/****************************************************************************/




/****************************************************************************/
/*                                                                          */
/*                       Public (member) functions                         */
/*                                                                          */
/****************************************************************************/


/* Constructors and destructors ********************************************/

/*
        This function initializes the class, not the object. DO NOT EVER
instantiate a PAWindow object without initializing its class...
*/
void pawindow_init (void) {

    char *tmpvalue;

    if (reference_count++) return;

    /* Get font, since its size will determine the minimum height */
    class.font=XLoadQueryFont(display, cfgpool_getstring(pawmpool,"pawm_font"));
    if (!class.font) BUG();

    /* Compute the baseline */
    /* Get the centered position calculating (Avaliable height - text heigt)/2
       and after that, add the font ascent to obtain the baseline */
    class.baseline=(TITLEBAR_H-(class.font->ascent+class.font->descent))/2;
    class.baseline+=class.font->ascent;

    class.gc=XCreateGC(display, ROOT, 0, NULL);
    if (class.gc==None) BUG();

    XSetForeground(display, class.gc, BLACK);
    XSetFont(display, class.gc, class.font->fid);
    XSetGraphicsExposures(display, class.gc, False);

    class.color.passivebar=x_color("titlebar_inactive");
    class.color.activebar=x_color("titlebar_active");
    class.color.passivetext=x_color("text_inactive");
    class.color.activetext=x_color("text_active");
    class.color.textshadow=x_color("text_shadow");
    class.color.framelight=x_color("frame_light");
    class.color.framewhite=x_color("frame_white");
    class.color.framebase=x_color("frame_base_color");
    class.color.framedark=x_color("frame_dark");
    class.color.framedarker=x_color("frame_darker");


    /* Build the pointers */
    class.cursor.NW=XCreateFontCursor(display, NWCURSOR);
    if (class.cursor.NW==None) BUG();
    class.cursor.NE=XCreateFontCursor(display, NECURSOR);
    if (class.cursor.NE==None) BUG();
    class.cursor.SW=XCreateFontCursor(display, SWCURSOR);
    if (class.cursor.SW==None) BUG();
    class.cursor.SE=XCreateFontCursor(display, SECURSOR);
    if (class.cursor.SE==None) BUG();
    class.cursor.resize=XCreateFontCursor(display, RESIZECURSOR);
    if (class.cursor.resize==None) BUG();
    class.cursor.move=MOVECURSOR;
    if (class.cursor.move==None) BUG();
 
    /* Build the button faces, masks aren't used for buttons, by now ;)) */
    tmpvalue=cfgpool_getstring(pawmpool,"close_normal");
    class.face.close_normal=x_build_pixmap(tmpvalue);
    free(tmpvalue);
    if (class.face.close_normal==NULL) BUG();
    
    tmpvalue=cfgpool_getstring(pawmpool,"close_active");
    class.face.close_active=x_build_pixmap(tmpvalue);
    free(tmpvalue);
    if (class.face.close_active==NULL) BUG();
    
    tmpvalue=cfgpool_getstring(pawmpool,"close_pressed");
    class.face.close_pressed=x_build_pixmap(tmpvalue);
    free(tmpvalue);
    if (class.face.close_pressed==NULL) BUG();
    
    tmpvalue=cfgpool_getstring(pawmpool,"min_normal");
    class.face.iconz_normal=x_build_pixmap(tmpvalue);
    free(tmpvalue);
    if (class.face.iconz_normal==NULL) BUG();
    
    tmpvalue=cfgpool_getstring(pawmpool,"min_active");
    class.face.iconz_active=x_build_pixmap(tmpvalue);
    free(tmpvalue);
    if (class.face.iconz_active==NULL) BUG();
    
    tmpvalue=cfgpool_getstring(pawmpool,"min_pressed");
    class.face.iconz_pressed=x_build_pixmap(tmpvalue);
    free(tmpvalue);
    if (class.face.iconz_pressed==NULL) BUG();
    
    tmpvalue=cfgpool_getstring(pawmpool,"max_normal");
    class.face.maximize_normal=x_build_pixmap(tmpvalue);
    free(tmpvalue);
    if (class.face.maximize_normal==NULL) BUG();
    
    tmpvalue=cfgpool_getstring(pawmpool,"max_active");
    class.face.maximize_active=x_build_pixmap(tmpvalue);
    free(tmpvalue);
    if (class.face.maximize_active==NULL) BUG();
    
    tmpvalue=cfgpool_getstring(pawmpool,"max_pressed");
    class.face.maximize_pressed=x_build_pixmap(tmpvalue);
    free(tmpvalue);
    if (class.face.maximize_pressed==NULL) BUG();
    

    /* Check shadow */ 
    if (strcmp(cfgpool_getstring(pawmpool,"window_shadow"),"On")==0)
	class.shadow_status=True;
    else
	class.shadow_status=False;

    /* No window has the focus ;) */
    focuspawindow=NULL;

    class.timestamp=0;
    class.clickx=-1;
    class.clicky=-1;	

    /* Initialize other needed classes */
    button_init();

    class.button.iconz=button_create(ROOT, 0, 0, BUTTONSZ, BUTTONSZ,
                                    class.face.iconz_normal->pixmap,
                                    class.face.iconz_active->pixmap,
                                    class.face.iconz_pressed->pixmap,
                                    iconify_client, NULL);


    class.button.close=button_create(ROOT, 0, 0, BUTTONSZ, BUTTONSZ,
                                     class.face.close_normal->pixmap,
                                     class.face.close_active->pixmap,
                                     class.face.close_pressed->pixmap,
                                     kill_client, NULL);

    class.button.maximize=button_create(ROOT, 0, 0, BUTTONSZ, BUTTONSZ,
		    			class.face.maximize_normal->pixmap,
					class.face.maximize_active->pixmap,
					class.face.maximize_pressed->pixmap,
					NULL, NULL);

    class.pawindows=chain_create();
    if (!class.pawindows) BUG();

    class.fpawindows=chain_create();
    if (!class.fpawindows) BUG();

}




/* This one destroys the class :(((( */
void pawindow_done (void) {

    PAWindow pawindow;	

    if (--reference_count) return;

    XFreeCursor(display, class.cursor.NW);
    XFreeCursor(display, class.cursor.NE);
    XFreeCursor(display, class.cursor.SW);
    XFreeCursor(display, class.cursor.SE);
    XFreeCursor(display, class.cursor.resize);

    XFreeFont(display, class.font);
    XFreeGC(display, class.gc);
    x_destroy_pixmap(class.face.close_normal);
    x_destroy_pixmap(class.face.close_active);
    x_destroy_pixmap(class.face.close_pressed);
    x_destroy_pixmap(class.face.iconz_normal);
    x_destroy_pixmap(class.face.iconz_active);
    x_destroy_pixmap(class.face.iconz_pressed);
    x_destroy_pixmap(class.face.maximize_normal);
    x_destroy_pixmap(class.face.maximize_active);
    x_destroy_pixmap(class.face.maximize_pressed);
    
    while ( (pawindow=chain_pop(class.fpawindows))!=NULL ); 
    
    chain_seek(class.pawindows, chain_size(class.pawindows));
    while ( (pawindow=chain_del(class.pawindows))!=NULL ) {
	pawindow_destroy(pawindow);
        XSync(display, False);
    }
    
    chain_delete(class.fpawindows);
    chain_delete(class.pawindows);

    button_destroy(class.button.close);
    button_destroy(class.button.iconz);
    button_destroy(class.button.maximize);
    
}




/* Makes a new pawindow from the window, but DOESN'T map it. */
PAWindow pawindow_create (Window window) {

    PAWindow pawindow;
    XTextProperty WMName;
    XWindowAttributes attribs;
    XSizeHints *szHints=NULL;
    long mask=0;
    Window w=None;
    Bool bound=False,clip=False;
    int foo;
        
    ASSERT(window);
    
    /** Basic initialization **/
    pawindow=malloc(PAWindowBytes);
    if (!pawindow) die("Not enough memory");
    
    /* Register newly created window */
    XAddToSaveSet(display, window);
    chain_seek(class.pawindows, chain_size(class.pawindows));
    chain_put(class.pawindows, pawindow);

    pawindow->window=window;
    pawindow->title=NULL;

    pawindow->transient=False;
    pawindow->active=False;
    pawindow->state=PA_WITHDRAWN;      /* ICCCM */
    pawindow->shaded=False;

    /* Does it need decoration? */
    pawindow->decorated=gtkHints_needDecoration(pawindow->window);

    /* Does it have a shape? */
    XShapeQueryExtents(display,pawindow->window,
	               &bound,&foo,&foo,&foo,&foo,
	               &clip,&foo,&foo,&foo,&foo);
    pawindow->shaped=bound;

    /* See if the window is transient or not */
    XGetTransientForHint(display, window, &w);
    pawindow->transient=(w==None) ? False : True;

    szHints=XAllocSizeHints();  /* Absolutely stupid... */
    if (!szHints) BUG();
        
    XSetWindowBorderWidth(display, window, 0);  /* NO border */

    pawindow->oldx=0; pawindow->oldy=0;
    pawindow->oldw=0; pawindow->oldh=0;

    /* Calculate sane values for geometry */
    if (XGetWindowAttributes(display, window, &attribs)) {
        if (!attribs.x && !attribs.y) {
	    /* If not default values are given, make the pawindows 'tile' to
	       make them more visible and not to be overlapped each other */
	    pawindow->x=(chain_size(class.pawindows)*10)-10;
	    pawindow->y=(chain_size(class.pawindows)*10)-10;
	} else {
	    pawindow->x=attribs.x;
	    pawindow->y=attribs.y;
	}
        pawindow->w=attribs.width;
        pawindow->h=attribs.height;
    } else {
	/* Set default values */    
        pawindow->x=(chain_size(class.pawindows)*10)-10;
        pawindow->y=(chain_size(class.pawindows)*10)-10;
        pawindow->w=FRAME_MIN_W;
        pawindow->h=FRAME_MIN_H;
    }

    pawindow->min_w=FRAME_MIN_W;
    pawindow->min_h=FRAME_MIN_H;
    pawindow->max_w=FRAME_MAX_W;
    pawindow->max_h=FRAME_MAX_H;


    /* Translate to the frame world */
    if (pawindow->decorated) {
        pawindow->w+=BORDERW*2;
        pawindow->h+=BORDERW*2+TITLEBAR_H;
    }

     /* Make titlebar visible at creation time */
    if (pawindow->x < 0) pawindow->x=0;
    if (pawindow->y < 0) pawindow->y=0;

    /* Make width & height not bigger than display size at creation time */
    if (pawindow->w > DWIDTH)  pawindow->w=DWIDTH;
    if (pawindow->h > DHEIGHT) pawindow->h=DHEIGHT;

    /* Another check */
    if ( pawindow->w > (DWIDTH-pawindow->x) )  pawindow->x=0;
    if ( pawindow->h > (DHEIGHT-pawindow->y) ) pawindow->y=0;

    /* At this point, X and Y for the frame are fully calculated */


    /*
        NOTE: the PA WM calculates *ALL* geometries based upon
    the frame geometry, including the client geometry, etc...
    */

    if (!XGetWMNormalHints(display, window, szHints, &mask)) szHints->flags=0;

    /* See whether we have size limits (client relative) */
    if (szHints->flags & PMinSize) {
        pawindow->min_w=szHints->min_width;
        pawindow->min_h=szHints->min_height;
	
        if (pawindow->decorated) {
         if (pawindow->min_w < FRAME_MIN_W) pawindow->min_w=FRAME_MIN_W;
         if (pawindow->min_h < FRAME_MIN_H) pawindow->min_h=FRAME_MIN_H;
         if (pawindow->min_w > FRAME_MAX_W) pawindow->min_w=FRAME_MAX_W;
         if (pawindow->min_h > FRAME_MAX_H) pawindow->min_h=FRAME_MAX_H;
	}
    }

    if (szHints->flags & PMaxSize) {
        pawindow->max_w=szHints->max_width;
        pawindow->max_h=szHints->max_height;

        if (pawindow->max_w < FRAME_MIN_W) pawindow->max_w=FRAME_MIN_W;
        if (pawindow->max_h < FRAME_MIN_H) pawindow->max_h=FRAME_MIN_H;
        if (pawindow->max_w > FRAME_MAX_W) pawindow->max_w=FRAME_MAX_W;
        if (pawindow->max_h > FRAME_MAX_H) pawindow->max_h=FRAME_MAX_H;
    }

    XFree(szHints);

    /* Sanitize values */
    if (pawindow->max_w < pawindow->min_w) pawindow->max_w=pawindow->min_w;
    if (pawindow->max_h < pawindow->min_h) pawindow->max_h=pawindow->min_h;

    /** Create the frame and set up decorations **/
    /* The shadow is, obviously, under the frame */

    /* Shadow window is created only when there is decoration */
    if (pawindow->decorated && class.shadow_status) {
	pawindow->shadow=XCreateSimpleWindow(display, ROOT,
	                                     SHADOW_X, SHADOW_Y,
                                             SHADOW_W, SHADOW_H,
                                             0, 0, BLACK);
	if (pawindow->shadow==None) BUG();
        XStoreName(display, pawindow->shadow, "shadow");
        XDefineCursor(display, pawindow->shadow, CURSOR);
    }
	
    
    /* The frame stores all decorations */
    pawindow->frame=XCreateSimpleWindow(display, ROOT,
                                        pawindow->x, pawindow->y,
                                        pawindow->w, pawindow->h,
                                        0, BLACK, BLACK);
    if (pawindow->frame==None) BUG();
    XStoreName(display, pawindow->frame, "frame");
    XDefineCursor(display, pawindow->frame, class.cursor.resize);


    /* The canvas just serve as a nest for the client */
    if (pawindow->decorated) {
        pawindow->canvas=XCreateSimpleWindow(display, pawindow->frame,
                                             CANVAS_X, CANVAS_Y,
                                             CANVAS_W, CANVAS_H,
                                             0, BLACK, BLACK);
    } else {
        pawindow->canvas=XCreateSimpleWindow(display, pawindow->frame,
                                             0, 0,
                                             pawindow->w, pawindow->h,
                                             0, BLACK, BLACK);
    }
    if (pawindow->canvas==None) BUG();

    XStoreName(display, pawindow->canvas, "canvas");
    XDefineCursor(display, pawindow->canvas, CURSOR);

    if (pawindow->decorated) {
        /* The titlebar... */
        pawindow->titlebar=XCreateSimpleWindow(display, pawindow->frame,
                                               TITLEBAR_X, TITLEBAR_Y,
                                               TITLEBAR_W, TITLEBAR_H,
                                               0, BLACK,
                                               class.color.passivebar);
        if (pawindow->titlebar==None) BUG();
        XStoreName(display, pawindow->titlebar, "titlebar");
        XDefineCursor(display, pawindow->titlebar, CURSOR);
    }
    
    /* Select events and register the Sloots */
    XSelectInput(display, pawindow->canvas, CANVAS_EVMASK);
    /* I love to use libraries with NO fucking documentation (ok, you
     * can found a scarce 7 page pdf and the man pages with the prototypes
     * for XShape, but that's all), another extensions (Xft comes to mind) 
     * may be in the same state... */
    XShapeSelectInput(display, pawindow->canvas, ShapeNotifyMask);
    if (!pawindow->decorated && pawindow->shaped) 
	set_shape(pawindow);    
    x_addsloot(pawindow->canvas, canvas_handle, pawindow);

    /* This handlers are only for decorated windows */
    if (pawindow->decorated) {
        XSelectInput(display, pawindow->titlebar, TITLEBAR_EVMASK);
        x_addsloot(pawindow->titlebar, titlebar_handle, pawindow);
    }
    
    /* This handler cannot be ignored for non-decorated windows, we
       need it to implement click-to-focus policy */
    XSelectInput(display, pawindow->frame, FRAME_EVMASK);
    x_addsloot(pawindow->frame, frame_handle, pawindow);

    /*  This is somewhat special, since we shouldn't install a
    handler for our client window ;)) and we just want one event...
    So, this event (PropertyNotify), will be handled by the 'canvas'
    */
    XSelectInput(display, pawindow->window, PropertyChangeMask);
    x_addsloot(pawindow->window, canvas_handle, pawindow);

    /* Reparent the client. Some stupid clients insist on being
    mapped out of screen and be here even if reparented. Well, we
    insist on moving them to the origin }:)))).  
    */
    XReparentWindow(display, window, pawindow->canvas, 0, 0);

    pawindow->icon=palabel_create(200, restore_client, pawindow);
    
    /* Get WM names, if available */
    if (XGetWMName(display, window, &WMName)) {
	SAY("Setting window name hint");
        set_title(pawindow, &WMName);
        XFree(WMName.value);
    } else {
	SAY("NULL title");
        set_title(pawindow, NULL);
    }

    if (XGetWMIconName(display, window, &WMName)) {
        palabel_settext(pawindow->icon, WMName.value, WMName.nitems);
        XFree(WMName.value);
    } else palabel_settext(pawindow->icon,
                           pawindow->title,
                           (pawindow->title)?strlen(pawindow->title):0);

    return pawindow;
}




/* This entirely destroys the specified pawindow */
void pawindow_destroy (PAWindow pawindow) {

    PAWindow pw;

    ASSERT(pawindow);

    SAY("Someone wants to destroy %#lx", pawindow->window);
    if (pawindow->title) free(pawindow->title);

    if (pawindow->decorated) {
        x_delsloot(pawindow->frame, frame_handle, pawindow);
        x_delsloot(pawindow->titlebar, titlebar_handle, pawindow);
        XSelectInput(display, pawindow->frame, NoEventMask);
        XSelectInput(display, pawindow->titlebar, NoEventMask);
    }
    x_delsloot(pawindow->canvas, canvas_handle, pawindow);
    x_delsloot(pawindow->window, canvas_handle, pawindow);

    /* No more events */
    XSelectInput(display, pawindow->canvas, NoEventMask);

    /* Give back window client to the root */
    XReparentWindow(display, pawindow->window, ROOT, pawindow->x, pawindow->y);
    XDestroyWindow(display, pawindow->frame);
    if (pawindow->decorated && class.shadow_status)
        XDestroyWindow(display, pawindow->shadow);
    pabar_dellabel(pawindow->icon);
    palabel_destroy(pawindow->icon);
   
    /* Delete pawindow from chains */
    chain_home(class.pawindows);
    while ( (pw=chain_next(class.pawindows))!=NULL ) {
        if (pw==pawindow) {
     	    chain_del(class.pawindows); 
	    break; 
        }
    }

    chain_home(class.fpawindows);
    while( (pw=chain_next(class.fpawindows))!=NULL ) {
	if (pw==pawindow) {
	    chain_del(class.fpawindows);
	}
    }
    
    /* Helps debugging... */
    memset(pawindow, '\0', PAWindowBytes);
    free(pawindow);

    /* Make sure nobody changed focuspawindow before. Damn global variables.*/
    if (!chain_size(class.fpawindows)) 
	focuspawindow=NULL;


}

/****************************************************************************/





/* Map handling functions *************************************************/

/* This maps and raises the pawindow and its decorations */
void pawindow_map (PAWindow pawindow) {

    CARD32 data[2];   /* Fucking X portable types */
              
    ASSERT(pawindow);
	
    SAY("Someone wants to map %#lx", pawindow->window);
    pawindow->active=True;

    if (!pawindow->decorated && pawindow->shaped)
       	set_shape(pawindow);

    if (pawindow->state==PA_WITHDRAWN) 
        pabar_addlabel(pawindow->icon);
	 
    /* Reparent and map buttons _before_ the pawindow, to make buttons draw
     * faster when the window is mapped */   
    if (pawindow->decorated) {

        button_reparent(class.button.close, pawindow->titlebar,
                        (int) TITLEBAR_W-BUTTONSZ, 0);
        button_action(class.button.close, kill_client, pawindow);
        button_reparent(class.button.maximize, pawindow->titlebar,
 	   	        (int) TITLEBAR_W-(BUTTONSZ*2), 0);
        button_action(class.button.maximize, maximize_client, pawindow);
        button_reparent(class.button.iconz, pawindow->titlebar,
                        (int) TITLEBAR_W-(BUTTONSZ*3), 0);
        button_action(class.button.iconz, iconify_client, pawindow);
	
        /* We paint the buttons except for transients */
        if (pawindow->transient==False) {
            button_map(class.button.close);
            button_map(class.button.iconz);
	    button_map(class.button.maximize);
        }
    } else {  /* No decoration, so buttons need to be put at a safe place ;) */
        button_unmap(class.button.close);
        button_unmap(class.button.iconz);
	button_unmap(class.button.maximize);
        button_reparent(class.button.close, ROOT, 0, 0);
        button_reparent(class.button.iconz, ROOT, 0, 0);
	button_reparent(class.button.maximize, ROOT, 0, 0);
    }

    
    /* Map the shadow window first, and resize it after mapping the client */
    if (pawindow->decorated && class.shadow_status) {
        XResizeWindow(display, pawindow->shadow, 1, 1);
        XMapRaised(display, pawindow->shadow);
    }
 
    if (!pawindow->shaded)   
        XMapSubwindows(display, pawindow->frame);
    else {
	XMapWindow(display, pawindow->frame);
	XMapWindow(display, pawindow->titlebar);
    }
    /* Map window two times, for slow drawing apps. First is the real map, 
     * second time only is raised */
    if (!pawindow->shaded)
        XMapWindow(display, pawindow->window);
    XMapRaised(display, pawindow->frame);
    if (!pawindow->shaded)
        XMapRaised(display, pawindow->window);

    if (pawindow->decorated && class.shadow_status) 
	XResizeWindow(display, pawindow->shadow, SHADOW_W, SHADOW_H);

    draw_border(pawindow);
    draw_titlebar(pawindow);

    /* Raised the pabar if needed */
    pabar_map();
    
    pawindow->state=PA_MAPPED;
    change_focus(pawindow);

    /* ICCCM, set the WM_STATE property */
    data[0]=NormalState;
    data[1]=None;
    XChangeProperty(display, pawindow->window, wm_state, wm_state,
                    32, PropModeReplace, (unsigned char *)data, 2);

}




/* Unmaps the pawindow, no tricks. By now... }:)  */
void pawindow_unmap (PAWindow pawindow) {

    CARD32 data[2];    /* Fucking X portable types */
    PAWindow pw;

    ASSERT(pawindow);

    SAY("Someone wants to unmap %#lx", pawindow->window);
    pawindow->state=PA_ICONIFIED;

    /* Ditto */
    if (pawindow->decorated && class.shadow_status)
        XUnmapWindow(display, pawindow->shadow);
    
    XUnmapSubwindows(display, pawindow->frame);
    XUnmapWindow(display, pawindow->frame);

    if (pawindow->active==False) return;

    button_unmap(class.button.close);
    button_unmap(class.button.iconz);
    button_unmap(class.button.maximize);
    button_reparent(class.button.close, ROOT, 0, 0);
    button_reparent(class.button.iconz, ROOT, 0, 0);
    button_reparent(class.button.maximize, ROOT, 0, 0);
    
    if (pawindow->icon) palabel_unsetfocus(pawindow->icon);
    
    /* Get a new pawindow to be raised */
    /* Remove all the ocurrencies of the unmapped pawindow and
       windows in withdrawn state from the focus chain */
    chain_home(class.fpawindows);
    while( (pw=chain_next(class.fpawindows))!=NULL ) {
	if (pw==pawindow || pw->state==PA_WITHDRAWN) 
	    chain_del(class.fpawindows);
    }

    /* Get the first pawindow in the stack */
    pw=chain_pop(class.fpawindows);
	
    if (pw!=NULL) pawindow_map(pw);
    else focuspawindow=NULL;
    
    /* ICCCM, set the WM_STATE property */
    data[0]=IconicState;
    data[1]=None;
    XChangeProperty(display, pawindow->window, wm_state, wm_state,
                    32, PropModeReplace, (unsigned char *)data, 2);

}



/****************************************************************************/






/* Geometry management procedures XDDD *************************************/

/* Resize a PAWindow. Notice that 'width' and 'height' are client relative*/
void pawindow_resize (PAWindow pawindow, unsigned int w, unsigned int h) 
{

    ASSERT(pawindow);

    /* Honor client size limits, but don't give a shit for shaded windows */
    w=(w < pawindow->min_w) ? pawindow->min_w : w;
    w=(w > pawindow->max_w) ? pawindow->max_w : w;
    h=(h < pawindow->min_h) ? pawindow->min_h : h;
    h=(h > pawindow->max_h) ? pawindow->max_h : h;

    /* Change titlebar size */
    if (pawindow->decorated)
        XResizeWindow(display, pawindow->titlebar, w, TITLEBAR_H);

    /* Displace buttons with the bar resize, but only if the resized
     * pawindow has the focus */
    if (pawindow==focuspawindow) {
        button_move(class.button.close, (int) w-BUTTONSZ, 0);
        button_move(class.button.maximize, (int) w-(BUTTONSZ*2), 0);
        button_move(class.button.iconz, (int) w-(BUTTONSZ*3), 0);
    }

    /* Change client's size */
    XResizeWindow(display, pawindow->window, w, h);
    XResizeWindow(display, pawindow->canvas, w, h);

    if (pawindow->decorated) {
       w+=BORDERW*2;
       h+=BORDERW*2+TITLEBAR_H;
    }

    pawindow->w=w;
    pawindow->h=h;

    /* And the other windows too */
    if (pawindow->decorated && class.shadow_status)
        XResizeWindow(display, pawindow->shadow, pawindow->w, pawindow->h);
    
    XResizeWindow(display, pawindow->frame, pawindow->w, pawindow->h);

}




/* Move a PAWindow.*/
/* NOTE: 'x' and 'y' are supposed to be client relative... */
void pawindow_move (PAWindow pawindow, int x, int y) 
{
    XEvent e;
    int abs_x,abs_y,h,w;

    ASSERT(pawindow);

    if (pawindow->decorated) {
	abs_x=x-BORDERW; 
	abs_y=y-BORDERW-TITLEBAR_H;
    	w=pawindow->w-BORDERW*2;
	h=pawindow->h-BORDERW*2-TITLEBAR_H;
    } else { 
        abs_x=x; abs_y=y;
	w=pawindow->w; h=pawindow->h;
    }

    /* Sanitize values if the window is being moved out of sight */
    if (abs_x < -w ) abs_x=-w+1;    
    if (abs_y < 1) abs_y=1;

    /* Build and send the notice for our client - ICCCM */
    /* Watch the fuck out! If the pawindow is shaded, only the titlebar
       is visible, so we DON'T want to send and ConfigureNotify to the client 
       or strange things are going to happen */
    if (!pawindow->shaded) {
        e.type=ConfigureNotify;
        e.xconfigure.event=pawindow->window;
        e.xconfigure.window=pawindow->window;
        e.xconfigure.x=abs_x;
        e.xconfigure.y=abs_y;
        e.xconfigure.width = w;
        e.xconfigure.height= h;
        e.xconfigure.border_width=0;
        e.xconfigure.above=None;
        e.xconfigure.override_redirect=False;

        if (!XSendEvent(display, pawindow->window, False,
                        StructureNotifyMask, &e)) BUG();
    }
		    
    /* We musn't change pawindow coordinates unless they're different 
     * from the new ones */		    
    if ( pawindow->x != x ) pawindow->x=abs_x;
    if ( pawindow->y != y ) pawindow->y=abs_y;

    XMoveWindow(display, pawindow->frame, pawindow->x, pawindow->y);
    
    if (pawindow->decorated && class.shadow_status)
        XMoveWindow(display, pawindow->shadow, SHADOW_X, SHADOW_Y);

}


/* Shade (Make only visible the titlebar) a pawindow*/
void pawindow_shade (PAWindow pawindow) 
{
    ASSERT(pawindow);

    pawindow->shaded=True;
    pawindow->oldsh=pawindow->h;
    pawindow->h=TITLEBAR_H+(2*BORDERW);
    
    /* Unmap client windows*/
    XUnmapWindow(display, pawindow->window);
    XUnmapWindow(display, pawindow->canvas);

    /* Resize shadow */
    if (class.shadow_status)
        XResizeWindow(display, pawindow->shadow, pawindow->w, pawindow->h);
    
    /* And finally resize the frame */
    XResizeWindow(display, pawindow->frame, pawindow->w, pawindow->h);

}


/* Unshade a pawindow*/
void pawindow_unshade (PAWindow pawindow) 
{
    ASSERT(pawindow);
    
    pawindow->shaded=False;
    pawindow->h=pawindow->oldsh;
 
    /* Resize and map again the subwindows */   
    XResizeWindow(display, pawindow->frame, pawindow->w, pawindow->h);

    XMapWindow(display, pawindow->canvas);
    XMapWindow(display, pawindow->window);

    if (class.shadow_status)
        XResizeWindow(display, pawindow->shadow, pawindow->w, pawindow->h);

    change_focus(pawindow);
    
}

/****************************************************************************/




/* Accessors****************************************************************/

/* This function returns a reference to the PAWindow icon */
PALabel pawindow_icon (PAWindow pawindow) {

    ASSERT(pawindow);
    ASSERT(pawindow->icon);
    
    return pawindow->icon;
}
/****************************************************************************/
