/* *** timux.c *** Simple Timer Multiplexor.

              Copyright (C) 2001 Ral Nez de Arenas Coronado

        This program is free software; you can redistribute it and/or
         modify it under the terms of the GNU General Public License
                as published by the Free Software Foundation;
                      either version 2 of the License,
                   or (at your option) any later version.

       This program is distributed in the hope that it will be useful,
         but WITHOUT ANY WARRANTY; without even the implied warranty
           of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
            See the GNU General Public License for more details.

      You should have received a copy of the GNU General Public License
             (`GPL') along with this program; if not, write to:

                       Free Software Foundation, Inc.
                         59 Temple Place, Suite 330
                         Boston, MA 02111-1307  USA

                   Report bugs to <raul@pleyades.net>
*/
#include <signal.h>
#include <sys/time.h>
#include "chain.h"
#include "timux.h"
#ifndef NDEBUG
 #define NDEBUG
#endif
#include "mobs.h"

typedef jobid TSC;

typedef struct TiMuxEvent {
        TSC timestamp;
        duty action;
        void *params;
} *TiMuxEvent;


struct TiMux {
    Chain events;
    TiMuxEvent target;
};

static volatile uintmax_t timux_count=0;
static TiMux theTiMux=NULL;

static void TheHandlerTM   (int);
static void TheSchedulerTM (void);




/*
    This is the handler for SIGALRM.
*/
static void TheHandlerTM (int dummy) {

    TiMuxEvent target;

    ASSERT(dummy==SIGALRM);
    ASSERT(theTiMux);
    ASSERT(theTiMux->target);
    
    dummy=errno;

    target=theTiMux->target;

    /* Run the job */
    SAY("About to run target %p (%#llx)\n", target, target->timestamp);

    if (target) target->action(target->params);

    free(target);

    theTiMux->target=NULL;

    /* Just to be sure, schedule again */
    TheSchedulerTM();
    errno=dummy;
}




/* This function gets the next target to run and schedules the timer */
static void TheSchedulerTM (void) {

    struct itimerval when={{0,0},{0,0}};
    TSC now=0;

    ASSERT(theTiMux);

    /* If the target is already defined, we are dry-running */
    if (theTiMux->target) {
        SAY("Already targeted");
        return;
    }
    
    /* Get the first one, if any */
    chain_seek(theTiMux->events, 1);
    theTiMux->target=chain_del(theTiMux->events);
    
    if (theTiMux->target) {
        /* Program the interval timer for the target... */
        if (gettimeofday(&(when.it_value), NULL) < 0) BUG();
        now+=(unsigned) when.it_value.tv_sec  * 1000;
        now+=(unsigned) when.it_value.tv_usec / 1000;

        SAY("Target is %p", theTiMux->target);    
        if (now >= theTiMux->target->timestamp) {
            SAY("Too fast");
            when.it_value.tv_sec=0;
            when.it_value.tv_usec=1;
        } else {
            SAY("now is %llu", now);
            SAY("tsc is %llu", theTiMux->target->timestamp);
            now=theTiMux->target->timestamp-now;
        
            /* Compute microseconds */
            when.it_value.tv_usec= (now % 1000) * 1000;  /* Always <= 1e6 */
            when.it_value.tv_sec= now / 1000;
        }
    }
    SAY("scheduled timer to %ld.%ld", when.it_value.tv_sec, when.it_value.tv_usec);
    setitimer(ITIMER_REAL, &when, NULL);
}




TiMux timux_create (void) {

    if (!timux_count) {
        struct sigaction action;
        struct itimerval dummy;

        theTiMux=malloc(sizeof(struct TiMux));
        if (!theTiMux) return NULL;

        /* Initialize event queue */
        theTiMux->events=chain_create();
        if (!(theTiMux->events)) {
            free(theTiMux);
            theTiMux=NULL;
            return NULL;
        }
	theTiMux->target=NULL;
	
        /* The signal need to be ignored to timux to work */
        sigaction(SIGALRM, NULL, &action);
        ASSERT(action.sa_handler==SIG_IGN);

        /* Stop any pending alarm */
        memset(&dummy, 0, sizeof(dummy));
        setitimer(ITIMER_REAL, &dummy, NULL);
        setitimer(ITIMER_VIRTUAL, &dummy, NULL);
        setitimer(ITIMER_PROF, &dummy, NULL);

        /* Establish our own handler */
        action.sa_handler=TheHandlerTM;
        action.sa_flags=SA_RESTART;
        sigfillset(&action.sa_mask);
        sigaction(SIGALRM, &action, NULL);
    }
    timux_count++;

    return theTiMux;
}




TiMux timux_delete (TiMux timux) {

    ASSERT(timux);

    timux_count--;

    if (!timux_count) {
        struct itimerval dummy;
        struct sigaction action;

        /* Stop any pending alarm */
        memset(&dummy, 0, sizeof(dummy));
        setitimer(ITIMER_REAL, &dummy, NULL);
        setitimer(ITIMER_VIRTUAL, &dummy, NULL);
        setitimer(ITIMER_PROF, &dummy, NULL);

        /* Discard any pending event to be executed */
        chain_seek(timux->events, 1);
        while (chain_del(timux->events));
        timux->events=chain_delete(timux->events);
        free(timux);
        theTiMux=NULL;

        action.sa_handler=SIG_IGN;
        action.sa_flags=SA_RESTART;
        sigfillset(&action.sa_mask);
        sigaction(SIGALRM, &action, NULL);
    }

    return NULL;
}




/*
    This method adds and schedules a new timed event, specified by 'action'
and 'params'. The 'when' parameter indicates how many milliseconds from now
must pass until the job is run. Please note that this time can be subtly
adjusted for proper operation and for avoiding job overlapping.

    The return value is the event indentifier, and marks uniquely this
timed event. This must be treated as an opaque type, since it depends
on the desired resolution and may change in the future.
    
    ENOMEM: Not enough memory to add the job.
*/
jobid timux_addjob (TiMux timux, duty action, void *params, tempus when) {

    TiMuxEvent job=NULL;
    sigset_t mask, oldmask;
    struct timeval now;

    ASSERT(timux);
    ASSERT(action);

    /* Compute the timestamp */
    if (gettimeofday(&now, NULL) < 0) BUG();

    job=malloc(sizeof(struct TiMuxEvent));
    if (!job) {
        errno=ENOMEM;
        return 0;
    }
    job->action=action;
    job->params=params;
    job->timestamp=when;
    job->timestamp+=(unsigned) now.tv_sec  * 1000;
    job->timestamp+=(unsigned) now.tv_usec / 1000;
    job->timestamp++;  /* Round up to the next millisecond */
    

    /* First of all, block the handler */
    sigemptyset(&mask);
    sigaddset(&mask, SIGALRM);
    sigprocmask(SIG_BLOCK, &mask, &oldmask);

    /* From now on the handler CANNOT execute */    

    /* Store back the current scheduled job */
    errno=0;
    chain_home(timux->events);
    if (timux->target) chain_put(timux->events, timux->target);
    if (errno) return 0;  /* Currently 'errno' must be ENOMEM */

    timux->target=NULL;

    /* Store the new job */
    chain_home(timux->events);
    while (1) {   /* Adjust the timestamp */
        TiMuxEvent tmp=chain_next(timux->events);

        if (!tmp) break;
        
        if (job->timestamp == tmp->timestamp) job->timestamp++;
        
        if (job->timestamp > tmp->timestamp) continue;

        /* Fallback: found insertion point */
        break;
    }

    /* Bingo!, we are at the appropriate place :)) */
    chain_prev(timux->events);
    errno=0;
    chain_put(timux->events, job);
    if (errno) {
        free(job);
        job=NULL;
        errno=0;
    }

    TheSchedulerTM();
    
    /* Leave whatever mask we had on entry */
    sigprocmask(SIG_SETMASK, &oldmask, NULL);

    /* The timestamp also serves as the job identifier */
    if (job) return job->timestamp;
    else { errno=ENOMEM; return 0; }
}




void timux_deljob (TiMux timux, jobid job) {

    sigset_t mask, oldmask;
    TSC timestamp=job;  /* Remember, 'TSC' is just like 'jobid' */

    ASSERT(timux);

    /* First of all, block the handler */
    sigemptyset(&mask);
    sigaddset(&mask, SIGALRM);
    sigprocmask(SIG_BLOCK, &mask, &oldmask);

    /* From now on the handler CANNOT execute */    

    /* Find the job to detach: is the current target? */
    if (timux->target->timestamp==timestamp) {
        free(timux->target);
        timux->target=NULL;
        TheSchedulerTM();  /* We need to run the scheduler... */
    } else {
        /* Is not the current target. Find it */
        chain_home(timux->events);
        while (1) {
            TiMuxEvent tmp=chain_next(timux->events);

            if (!tmp) break;  /* Empty event queue */

            if (tmp->timestamp==timestamp) {  /* Got it :))) */
                free(chain_del(timux->events));
                break;
            }
            
        }
    }

    /* Leave whatever mask we had on entry */
    sigprocmask(SIG_SETMASK, &oldmask, NULL);

}
